import pandas as pd
import datetime as dt


def create_calibration_dataframe_alh():
    # Dictionary organized by wavelengths
    depoCalib = {"ff": {}, "nf": {}}

    depoCalib["ff"]["355"] = pd.DataFrame(
        columns=["date", "eta_an", "eta_pc", "GR", "GT", "HR", "HT", "K"]
    )
    depoCalib["nf"]["355"] = pd.DataFrame(
        columns=["date", "eta_an", "eta_pc", "GR", "GT", "HR", "HT", "K"]
    )
    depoCalib["nf"]["532"] = pd.DataFrame(
        columns=["date", "eta_an", "eta_pc", "GR", "GT", "HR", "HT", "K"]
    )

    dict0 = {
        "date": dt.datetime(2023, 3, 2),
        "eta_an": 1.4664,
        "eta_pc": 1.2291,
        "GR": 2.28007,
        "GT": 2.36168,
        "HR": 2.27859,
        "HT": -2.36014,
        "K": 1.002,
    }  # Considering y = + 1 i.e., parallel in T.
    dict1 = {
        "date": dt.datetime(2023, 3, 2),
        "eta_an": 1.4383,
        "eta_pc": 1.336,
        "GR": 2.15778,
        "GT": 1.83852,
        "HR": 2.15241,
        "HT": -1.83393,
        "K": 1.002,
    }  # Considering y = - 1 i.e., parallel in R.
    depoCalib["ff"]["355"].loc[0] = pd.Series(dict1)
    depoCalib["nf"]["355"].loc[0] = pd.Series(dict1)
    depoCalib["nf"]["532"].loc[0] = pd.Series(dict0)
    depoCalib["ff"]["355"] = depoCalib["ff"]["355"].set_index("date")
    depoCalib["nf"]["355"] = depoCalib["nf"]["355"].set_index("date")
    depoCalib["nf"]["532"] = depoCalib["nf"]["532"].set_index("date")
    return depoCalib



from typing import Literal, Tuple

import matplotlib
import numpy as np
import xarray as xr
from loguru import logger
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from gfatpy.utils import plot

BoundsType = tuple[float, float] | Literal["auto", "limits"]


def get_norm(
    rcs: np.ndarray,
    scale_bounds: BoundsType,
    color_resolution: int = 128,
) -> matplotlib.colors.BoundaryNorm:

    match scale_bounds:
        case "auto":
            bounds = np.linspace(0, rcs.max() * 0.6, 128)
        case "limits":
            bounds = np.linspace(rcs.min(), rcs.max(), 128)
        case _:
            bounds = np.linspace(*scale_bounds, 128)

    norm = matplotlib.colors.BoundaryNorm(boundaries=bounds, ncolors=2**8, clip=False)
    logger.debug(f"Color bounds min - max: {bounds.min()} - {bounds.max()}")

    return norm


def apply_labels(ax: matplotlib.axes.Axes, data_array: xr.DataArray) -> None:  # type: ignore
    # TODO: Other titles will later be added with config param
    plot.title1(data_array.name, 2)
    plot.title2(np.atleast_1d(data_array.time.mean().dt.date.values)[0].isoformat(), 2)
    # plot.title3(
    #     "{} ({:.1f}N, {:.1f}E)".format(
    #         data_array.attrs.get("site_location"),
    #         float(data_array.attrs["geospatial_lat_min"]),
    #         float(data_array.attrs["geospatial_lon_min"]),
    #     ),
    #     2,
    # )

    plot.watermark(ax, zoom=0.6, alpha=0.6)


def apply_gap_size(ax: matplotlib.axes.Axes, data_array) -> None:  # type: ignore
    diff = data_array.time[1:].values - data_array.time[0:-1].values
    gap_size = 2 * int(
        np.ceil(
            np.median(np.median(diff).astype("timedelta64[s]").astype("float") / 60)
        )
    )

    plot.gapsizer(
        ax,
        data_array.time.values.astype("M8[ms]").astype("O"),
        data_array.range.values,
        gap_size,
        "#c7c7c7",
    )


# TODO: this should be the default implemetation rather than filelist argument
def quicklook_xarray(
    data_array: xr.DataArray,
    /,
    is_rcs: bool = True,
    scale_bounds: BoundsType = "auto",
    colormap: str | matplotlib.colors.Colormap = "jet",
) -> Tuple[Figure, Axes]:
    """Plot a quicklook of a xarray.DataArray

    Args:
        data_array (xr.DataArray): Lidar signal or rcs from `lidar.preprocess`
        is_rcs (bool, optional): To indicate if the data is RCS or not. Defaults to True.
        scale_bounds (BoundsType, optional): scale bounds for the colorbar. Defaults to "auto".
        colormap (str | matplotlib.colors.Colormap, optional): colormap of the colorbar. Defaults to "jet".

    Returns:
        tuple[Figure, Axes]: Figure and Axes objects
    """

    try:
        if is_rcs:
            rcs = data_array.values
        else:
            rcs = data_array.values * data_array.range.values**2

        fig, ax = plt.subplots(figsize=(15, 5))
        norm = get_norm(rcs, scale_bounds)

        q = ax.pcolormesh(
            data_array.time, data_array.range, rcs.T, cmap=colormap, norm=norm
        )

        ax.set_xlabel(r"Time, $[UTC]$")
        ax.set_ylabel(r"Height, $[m, \, agl]$")
        ax.set_ylim(0)

        apply_labels(ax, data_array=data_array)
        apply_gap_size(ax, data_array=data_array)

        fig.colorbar(q)

        q.cmap.set_over("white")  # type: ignore
    except Exception as e:
        raise e
    return fig, ax



from enum import Enum
from typing import TypedDict

from typing_extensions import NotRequired

import numpy as np


class RadarName(str, Enum):
    nph = "NEPHELE"
    nbl = "NEBULA"    

class MeasurementType(str, Enum):
    RS = "RS"
    HF = "HF"
    DC = "DC"
    TC = "TC"
    DP = "DP"
    OT = "OT"


class _RadarInfo(TypedDict):
    NEPHELE: dict
    NEBULA: dict

class _RadarMetadata(TypedDict):
    nick2name: dict
    name2nick: dict
    measurement_type: list

class RadarInfoType(TypedDict):
    radars: _RadarInfo
    metadata: _RadarMetadata


class ParamsDict(TypedDict):
    k_radar: float
    # particle_alpha: np.ndarray
    # particle_alpha_raman: NotRequired[np.ndarray]
    # particle_beta: np.ndarray
    # particle_beta_raman: NotRequired[np.ndarray]
    # molecular_alpha: np.ndarray
    # molecular_alpha_raman: NotRequired[np.ndarray]
    # molecular_beta: np.ndarray
    # molecular_beta_raman: NotRequired[np.ndarray]
    # particle_accum_ext: np.ndarray
    # particle_accum_ext_raman: NotRequired[np.ndarray]
    # molecular_accum_ext: np.ndarray
    # molecular_accum_ext_raman: NotRequired[np.ndarray]
    # molecular_beta_att: np.ndarray
    # transmittance: np.ndarray
    # overlap: np.ndarray
    # angstrom_exponent_fine: float
    # angstrom_exponent_coarse: float



from pdb import set_trace
import zipfile
import tempfile
import numpy as np
from typing import Any, Tuple
from pathlib import Path
from datetime import datetime

# from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
# from pdfminer.converter import TextConverter
# from pdfminer.layout import LAParams
# from pdfminer.pdfpage import PDFPage
from io import StringIO
import yaml


def read_yaml(path: Path | str) -> Any:
    """
    Reads a yaml file and returns the data as a dictionary.
    """
    with open(path, "r") as stream:
        return yaml.safe_load(stream)


# def convert_pdf_to_txt(path, pages=None):
#     if not pages:
#         pagenums = set()
#     else:
#         pagenums = set(pages)
#     output = StringIO()
#     manager = PDFResourceManager()
#     converter = TextConverter(manager, output, laparams=LAParams())
#     interpreter = PDFPageInterpreter(manager, converter)

#     infile = open(path, "rb")
#     for page in PDFPage.get_pages(infile, pagenums):
#         interpreter.process_page(page)
#     infile.close()
#     converter.close()
#     text = output.getvalue()
#     output.close()
#     return text


from pathlib import Path
import numpy as np
from datetime import datetime


def find_nearest_filepath(
    dir: Path, wildcard_filename: str, date_location_in_filename: int, date: datetime
) -> Path:
    """Finds the nearest file path in the specified directory based on a wildcard filename pattern and a target date.

    Args:
        dir (Path): The directory to search for files.
        wildcard_filename (str): The wildcard filename pattern to match files.
        date_location_in_filename (int): The index of the date in the filename split by underscores.
        date (datetime): The target date to find the nearest file.

    Returns:
        Path: The path to the nearest file.

    Raises:
        ValueError: If no file is found in the directory.

    """
    # Get all candidate file paths
    candidates = [*dir.rglob(wildcard_filename)]

    # Raise an error if no candidate files are found
    if not candidates:
        raise ValueError("No file found.")

    # Extract the dates from candidate file names
    candidate_dates = np.array(
        [
            np.datetime64(p.name.split(".")[0].split("_")[date_location_in_filename])
            for p in candidates
        ]
    )

    # Calculate the differences between target date and candidate dates
    date_diffs = np.abs(candidate_dates - np.datetime64(date))

    # Find the index of the minimum date difference
    idx = np.argmin(date_diffs)

    # Get the path of the nearest file
    path = candidates[idx]

    # Raise an error if the path does not exist
    if not path.exists():
        raise ValueError("No file found.")

    return path

def unzip_file_to_temp(file: Path) -> Path:
    
    if file.suffix != ".zip":
        raise ValueError("The file is not a zip file.")
    
    filename = file.name.split(".")[0]

    # Create a temporary directory    
    temp_dir = Path(tempfile.mkdtemp(prefix="tmp_unzipped_", dir=Path.cwd().as_posix()))
    unzip_file = temp_dir / filename
    try:
        # Extract the zip file
        with zipfile.ZipFile(file, 'r') as zip_ref:
            zip_ref.extractall(unzip_file)
            
    except zipfile.BadZipFile:
        raise ValueError("The file is not a valid zip file.")        

    except Exception as e:
        raise ValueError(f"An error occurred: {str(e)}")            
    return unzip_file            


import pandas as pd

from gfatpy.aeronet.utils import add_interpol_aod, AERONET_INFO


def aod_lidar_wavelengths_all(header: dict, df: pd.DataFrame) -> pd.DataFrame:
    """Adds the AOD at 355, 532 and 1064 nm to the AERONET dataframe.

    Args:
        header (dict): AERONET header.
        df (pd.DataFrame): AERONET dataframe.

    Returns:
        pd.DataFrame: AERONET dataframe.
    """

    for wave_ in [355, 532, 1064]:
        aod_str = AERONET_INFO["aod_string_type"][
            header["aeronet_data"].replace(" ", "_")
        ].replace("&", f"{wave_}")
        aod_ = add_interpol_aod(
            header,
            df,
            wave_,
        )
        df[aod_str] = aod_

    return df



"""
Tomado prestado de lidar_processing.lidar_processing.helper_functions
"""
from typing import Any

import numpy as np
import pandas as pd
import xarray as xr
import scipy
from scipy import interpolate
from scipy.constants import Boltzmann as k_b
from pandas.core.arrays import ExtensionArray

# Scattering parameters according to Freudenthaler V., 2015.
Fk = {
    308: 1.05574,
    351: 1.05307,
    354.717: 1.05209,
    355: 1.05288,
    386.890: 1.05166,
    400: 1.05125,
    407.558: 1.05105,
    510.6: 1.04922,
    532: 1.04899,
    532.075: 1.04899,
    607.435: 1.04839,
    710: 1.04790,
    800: 1.04763,
    1064: 1.04721,
    1064.150: 1.04721,
}

epsilon = {
    308: 0.25083,
    351: 0.238815,
    354.717: 0.234405,
    355: 0.23796,
    386.890: 0.23247,
    400: 0.230625,
    407.558: 0.229725,
    510.6: 0.22149,
    532: 0.220455,
    532.075: 0.220455,
    607.435: 0.217755,
    710: 0.21555,
    800: 0.214335,
    1064: 0.212445,
    1064.150: 0.212445,
}

Cs = {
    308: 3.6506e-5,  # K/hPa/m
    351: 2.0934e-5,
    354.717: 2.0024e-5,
    355: 1.9957e-5,
    386.890: 1.3942e-5,
    400: 1.2109e-5,
    407.558: 1.1202e-5,
    510.6: 4.4221e-6,
    532: 3.7382e-6,
    532.075: 3.7361e-6,
    607.435: 2.1772e-6,
    710: 1.1561e-6,
    800: 7.1364e-7,
    1064: 2.2622e-7,
    1064.150: 2.2609e-7,
}

BsT = {
    308: 4.2886e-6,
    351: 2.4610e-6,
    354.717: 2.3542e-6,
    355: 2.3463e-6,
    400: 1.4242e-6,
    510.6: 5.2042e-7,
    532: 4.3997e-7,
    532.075: 4.3971e-7,
    710: 1.3611e-7,
    800: 8.4022e-8,
    1064: 2.6638e-8,
    1064.150: 2.6623e-8,
}

BsC = {
    308: 4.1678e-6,
    351: 2.3949e-6,
    354.717: 2.2912e-6,
    355: 2.2835e-6,
    400: 1.3872e-6,
    510.6: 5.0742e-7,
    532: 4.2903e-7,
    532.075: 4.2878e-7,
    710: 1.3280e-7,
    800: 8.1989e-8,
    1064: 2.5999e-8,
    1064.150: 2.5984e-8,
}

BsC_parallel = {
    308: 4.15052184e-6,
    351: 2.38547616e-06,
    354.717: 2.28368241e-06,
    355: 2.27451222e-06,
    400: 1.38198631e-06,
    510.6: 5.05563542e-07,
    532: 4.27459520e-07,
    532.075: 4.27219387e-07,
    710: 1.32322062e-07,
    800: 8.16989147e-08,
    1064: 2.59074156e-08,
    1064.150: 2.58925276e-08,
}

BsC_perpendicular = {
    308: 1.72550768e-08,
    351: 9.44466863e-09,
    354.717: 8.87554356e-09,
    355: 8.97326485e-09,
    400: 5.28492464e-09,
    510.6: 1.85714694e-09,
    532: 1.56293632e-09,
    532.075: 1.56205831e-09,
    710: 4.73100855e-10,
    800: 2.90465461e-10,
    1064: 9.13006514e-11,
    1064.150: 9.12481844e-11,
}

KbwT = {
    308: 1.01610,
    351: 1.01535,
    354.717: 1.01530,
    355: 1.01530,
    400: 1.01484,
    510.6: 1.01427,
    532: 1.01421,
    532.075: 1.01421,
    710: 1.01390,
    800: 1.01383,
    1064: 1.01371,
    1064.150: 1.01371,
}

KbwC = {
    308: 1.04554,
    351: 1.04338,
    354.717: 1.04324,
    355: 1.04323,
    400: 1.04191,
    510.6: 1.04026,
    532: 1.04007,
    532.075: 1.04007,
    710: 1.03919,
    800: 1.03897,
    1064: 1.03863,
    1064.150: 1.03863,
}

molecular_depolarizationC = {
    351: 0.004158,
    354.717: 0.003959,
    355: 0.003956,
    400: 0.003825,
    510.6: 0.003673,
    532: 0.003656,
    532.075: 0.003656,
    710: 0.003575,
    800: 0.003555,
    1064: 0.003524,
    1064.150: 0.003524,
}

molecular_depolarizationT = {
    351: 0.01559,
    354.717: 0.01554,
    355: 0.01554,
    400: 0.01507,
    510.6: 0.01448,
    532: 0.01441,
    532.075: 0.01441,
    710: 0.01410,
    800: 0.01402,
    1064: 0.01390,
    1064.150: 0.01390,
}

# Create interpolation function once, to avoid re-calculation (does it matter?)
f_ext = interpolate.interp1d(list(Cs.keys()), list(Cs.values()), kind="cubic")
f_bst = interpolate.interp1d(list(BsT.keys()), list(BsT.values()), kind="cubic")
f_bsc = interpolate.interp1d(list(BsC.keys()), list(BsC.values()), kind="cubic")
f_bsc_parallel = interpolate.interp1d(
    list(BsC_parallel.keys()), list(BsC_parallel.values()), kind="cubic"
)
f_bsc_perpendicular = interpolate.interp1d(
    list(BsC_perpendicular.keys()), list(BsC_perpendicular.values()), kind="cubic"
)

# Splines introduce arifacts due to limited input resolution
f_kbwt = interpolate.interp1d(list(KbwT.keys()), list(KbwT.values()), kind="linear")
f_kbwc = interpolate.interp1d(list(KbwC.keys()), list(KbwC.values()), kind="linear")
f_molt = interpolate.interp1d(
    list(molecular_depolarizationT.keys()),
    list(molecular_depolarizationT.values()),
    kind="linear",
)
f_molc = interpolate.interp1d(
    list(molecular_depolarizationC.keys()),
    list(molecular_depolarizationC.values()),
    kind="linear",
)


def _standard_atmosphere(
    altitude: float,
    temperature_surface: float = 288.15,
    pressure_surface: float = 101325.0,
) -> tuple[float, float, float]:
    """
    Calculation of Temperature and Pressure in Standard Atmosphere.

    Parameters
    ----------
    altitude: float
       The altitude above sea level. (m)

    Returns
    -------
    pressure: float
       The atmospheric pressure. (N * m^-2 or Pa)
    temperature: float
       The atmospheric temperature. (K)
    density: float
       The air density. (kg * m^-3)

    References
    ----------
    http://home.anadolu.edu.tr/~mcavcar/common/ISAweb.pdf
    """

    # Dry air specific gas constant. (J * kg^-1 * K^-1)
    R = 287.058

    g = 9.8  # m/s^2

    # Temperature calculation.
    if altitude < 11000:
        temperature = temperature_surface - 6.5 * altitude / 1000.0
    else:
        temperature = temperature_surface - 6.5 * 11000 / 1000.0
    # Pressure calculation.
    if altitude < 11000:
        pressure = (
            pressure_surface * (1 - (0.0065 * altitude / temperature_surface)) ** 5.2561
        )
    else:
        # pressure = pressure_surface*((temperature/scaled_T[idx])**-5.2199))\
        #                       *np.exp((-0.034164*(_height - z_tmp))/scaled_T[idx])
        tropopause_pressure = (
            pressure_surface * (1 - (0.0065 * 11000 / temperature_surface)) ** 5.2561
        )
        tropopause_temperature = temperature
        pressure = tropopause_pressure * np.exp(
            -(altitude - 11000) * (g / (R * tropopause_temperature))
        )

    # Density calculation.
    density = pressure / (R * temperature)

    return pressure, temperature, density


def standard_atmosphere(
    altitude: np.ndarray[Any, np.dtype[np.float64]],
    temperature_surface: float = 288.15,
    pressure_surface: float = 101325.0,
) -> tuple[
    np.ndarray[Any, np.dtype[np.float64]],
    np.ndarray[Any, np.dtype[np.float64]],
    np.ndarray[Any, np.dtype[np.float64]],
]:
    """Calculation of Temperature and Pressure Profiles in Standard Atmosphere.

    Args:
        altitude (np.ndarray[Any, np.dtype[np.float64]]): _description_
        temperature_surface (float, optional): _description_. Defaults to 288.15.
        pressure_surface (float, optional): _description_. Defaults to 101325.0.

    Returns:
        tuple[np.ndarray[Any, np.dtype[np.float64]], np.ndarray[Any, np.dtype[np.float64]], np.ndarray[Any, np.dtype[np.float64]]]: _description_
    """

    pressure, temperature, density = (
        np.ones(len(altitude)),
        np.ones(len(altitude)),
        np.ones(len(altitude)),
    )
    for idx, altitude_ in enumerate(altitude):
        pressure[idx], temperature[idx], density[idx] = _standard_atmosphere(
            altitude=altitude_,
            temperature_surface=temperature_surface,
            pressure_surface=pressure_surface,
        )
    return pressure, temperature, density


def generate_meteo_profiles(
    heights: np.ndarray[Any, np.dtype[np.float64]],
    pressure: float | np.ndarray[Any, np.dtype[np.float64]] | None = None,
    temperature: float | np.ndarray[Any, np.dtype[np.float64]] | None = None,
) -> pd.DataFrame:
    """If Pressure and Temperature are floats, it provides scaled standard atmosphere using P,T as surface values. If vectors, scaled standard atmosphere if used on the top to fullfil them up to the heights vector provided.

    Args:
        pressure (float | np.ndarray[Any, np.dtype[np.float64]]): pressure
        temperature (float | np.ndarray[Any, np.dtype[np.float64]]): temperature
        heights (np.ndarray[Any, np.dtype[np.float64]]): range vector

    Returns:
        pd.DataFrame: pressure and temperature data
    """

    P = pressure
    T = temperature

    if P is None and T is None:
        # standard atmosphere profile:
        extended_P, extended_T, _ = standard_atmosphere(heights)
    elif isinstance(P, float) and isinstance(T, float):
        # standard atmosphere profile:
        extended_P, extended_T, _ = standard_atmosphere(
            heights, pressure_surface=P, temperature_surface=T
        )
    else:
        # standard atmosphere profile:
        Psa, Tsa, _ = standard_atmosphere(heights)

        if P.size == Psa.size:  # if they are the same size, we leave them be
            extended_P = P
        elif (
            P.size > Psa.size
        ):  # If our pressure vector is bigger than 'heights', we make it the same size
            maxsa = Psa.size
            extended_P = np.ones(heights.size) * np.nan
            for i in range(maxsa):
                extended_P[i] = P[i]
        else:  # if we don't have enough data to make a full pressure profile
            extended_P = P
            maxh = P.size  # number of the last data
            for i in range(Psa.size - maxh):
                extended_P = np.append(
                    extended_P, Psa[maxh + i]
                )  # we use standard atmosphere as our pressure profile

        if T.size == Tsa.size:  # if they are the same size, we leave them be
            extended_T = T
        elif (
            T.size > Tsa.size
        ):  # If our temperature vector is bigger than 'heights', we make it the same size
            maxsa = Tsa.size
            extended_T = np.ones(heights.size) * np.nan
            for i in range(maxsa):
                extended_T[i] = T[i]
        else:  # if we don't have enough data to make a full temperature profile
            extended_T = T
            maxh = T.size  # number of the last data
            for i in range(Tsa.size - maxh):
                extended_T = np.append(
                    extended_T, Tsa[maxh + i]
                )  # we use standard atmosphere as our temperature profile

    atmospheric_profiles = pd.DataFrame(
        {
            "height": heights,
            "temperature": extended_T,
            "pressure": extended_P,
        }
    )

    return atmospheric_profiles


def transmittance(
    alpha: np.ndarray[Any, np.dtype[np.float64]],
    heights: np.ndarray[Any, np.dtype[np.float64]],
) -> np.ndarray[Any, np.dtype[np.float64]]:
    """
    Transmittance = exp[-integral{alpha*dz}_[0,z]]

    Parameters
    ----------
    alpha: array
        extinction profile
    heights: array
        heights profile

    Returns
    -------
    transmittance: array
         transmittance

    """

    delta_height = np.median(np.diff(heights))
    integrated_extinction = scipy.integrate.cumtrapz(alpha, initial=0, dx=delta_height)
    return np.exp(-integrated_extinction)


def molecular_backscatter(
    wavelength: float,
    pressure: np.ndarray[Any, np.dtype[np.float64]],
    temperature: np.ndarray[Any, np.dtype[np.float64]],
    component: str = "total",
) -> np.ndarray[Any, np.dtype[np.float64]]:
    """
    Molecular backscatter calculation.

    Parameters
    ----------
    wavelength : float
       The wavelength of the radiation in air. From 308 to 1064.15
    pressure : float
       The atmospheric pressure. (Pa)
    temperature : float
       The atmospheric temperature. (K)
    component : str
       One of 'total' or 'cabannes'.

    Returns
    -------
    beta_molecular: float
       The molecular backscatter coefficient. (m^-1 * sr^-1)

    References
    ----------
    Freudenthaler, V. Rayleigh scattering coefficients and linear depolarization
    ratios at several EARLINET lidar wavelengths. p.6-7 (2015)
    """
    if component not in [
        "total",
        "cabannes",
        "cabannes_parallel",
        "cabannes_perpendicular",
    ]:
        raise ValueError(
            "Molecular backscatter available only for 'total' or 'cabannes' component."
        )

    if component == "total":
        bs_function = f_bst
    elif component == "cabannes":
        bs_function = f_bsc
    elif component == "cabannes_parallel":
        bs_function = f_bsc_parallel
    elif component == "cabannes_perpendicular":
        bs_function = f_bsc_perpendicular
    else:
        raise ValueError(f"{component} not found.")

    Bs = bs_function(wavelength)

    # Convert pressure to correct units for calculation. (Pa to hPa)
    pressure = pressure / 100.0

    # Calculate the molecular backscatter coefficient.
    beta_molecular = Bs * pressure / temperature

    return beta_molecular


def molecular_lidar_ratio(wavelength: float, component: str = "total") -> float:
    """
    Molecular lidar ratio.

    Parameters
    ----------
    wavelength : float
       The wavelength of the radiation in air. From 308 to 1064.15
    component : str
       One of 'total' or 'cabannes'.

    Returns
    -------
    lidar_ratio_molecular : float
       The molecular backscatter coefficient. (m^-1 * sr^-1)

    References
    ----------
    Freudenthaler, V. Rayleigh scattering coefficients and linear depolarization
    ratios at several EARLINET lidar wavelengths. p.6-7 (2015)
    """
    if component not in ["total", "cabannes"]:
        raise ValueError(
            "Molecular lidar ratio available only for 'total' or 'cabannes' component."
        )

    if component == "total":
        k_function = f_kbwt
    else:
        k_function = f_kbwc

    Kbw = k_function(wavelength)

    lidar_ratio_molecular = 8 * np.pi / 3.0 * Kbw

    return lidar_ratio_molecular


def molecular_depolarization(wavelength: float, component: str = "total") -> float:
    """
    Molecular lidar ratio.

    Parameters
    ----------
    wavelength : float
       The wavelength of the radiation in air. From 308 to 1064.15
    component : str
       One of 'total' or 'cabannes'.

    Returns
    -------
    molecular volume depolarization: float

    References
    ----------
    Freudenthaler, V. Rayleigh scattering coefficients and linear depolarization
    ratios at several EARLINET lidar wavelengths. p.6-7 (2015)
    """

    if component not in ["total", "cabannes"]:
        raise ValueError(
            "Molecular lidar ratio available only for 'total' or 'cabannes' component."
        )

    if component == "total":
        moldepo_function = f_molt
    else:
        moldepo_function = f_molc

    molecular_depolarization = moldepo_function(wavelength)

    return molecular_depolarization


def molecular_extinction(
    wavelength: float,
    pressure: np.ndarray[Any, np.dtype[np.float64]],
    temperature: np.ndarray[Any, np.dtype[np.float64]],
) -> np.ndarray[Any, np.dtype[np.float64]]:
    """
    Molecular extinction calculation.

    Parameters
    ----------
    wavelength : float
       The wavelength of the radiation in air. From 308 to 1064.15
    pressure : float
       The atmospheric pressure. (Pa)
    temperature : float
       The atmospheric temperature. (K)

    Returns
    -------
    alpha_molecular: float
       The molecular extinction coefficient. (m^-1)

    References
    ----------
    Freudenthaler, V. Rayleigh scattering coefficients and linear depolarization
    ratios at several EARLINET lidar wavelengths. p.6-7 (2015)
    """
    cs = f_ext(wavelength)

    # Convert pressure to correct units for calculation. (Pa to hPa)
    pressure = pressure / 100.0

    # Calculate the molecular backscatter coefficient.
    alpha_molecular = cs * pressure / temperature

    return alpha_molecular


def attenuated_backscatter(
    backscatter: np.ndarray[Any, np.dtype[np.float64]],
    transmittance: np.ndarray[Any, np.dtype[np.float64]],
) -> np.ndarray[Any, np.dtype[np.float64]]:
    """Calculate Attenuated Backscatter

    Args:
        backscatter ([type]): [description]
        transmittance ([type]): [description]
    """

    return backscatter * transmittance**2


def number_density_at_pt(
    pressure: float | np.ndarray, temperature: float | np.ndarray
) -> np.ndarray[Any, np.dtype[np.float64]]:
    """Calculate the number density for a given temperature and pressure.

    This method does not take into account the compressibility of air.

    Parameters
    ----------
    pressure: float or array
       Pressure in Pa
    temperature: float or array
       Temperature in K

    Returns
    -------
    n: array or array
       Number density of the atmosphere [m-3]
    """
    # p_pa = pressure * 100.  # Pressure in pascal

    n = pressure / (temperature * k_b)

    return n


def molecular_properties(
    wavelength: float,
    pressure: np.ndarray[Any, np.dtype[np.float64]] | ExtensionArray,
    temperature: np.ndarray[Any, np.dtype[np.float64]] | ExtensionArray,
    heights: np.ndarray[Any, np.dtype[np.float64]] | ExtensionArray,
    times: np.ndarray[Any, np.dtype[np.float64]] | None = None,
    component: str = "total",
) -> xr.Dataset:
    """Molecular Attenuated  Backscatter: beta_mol_att = beta_mol * Transmittance**2

    Args:
        wavelength (float): wavelength of our desired beta molecular attenuated
        pressure (np.ndarray[Any, np.dtype[np.float64]]): pressure profile
        temperature (np.ndarray[Any, np.dtype[np.float64]]): temperature profile
        heights (np.ndarray[Any, np.dtype[np.float64]]): height profile
        times (np.ndarray[Any, np.dtype[np.float64]] | None, optional): time array. Defaults to None. Note: this is not used in the calculation yet. In development.
        component (str, optional): _description_. Defaults to "total".

    Returns:
        xr.Dataset: molecular attenuated backscatter profile, molecular backscatter profile, molecular extinction profile, molecular lidar ratio, molecular depolarization ratio
    """

    # molecular backscatter and extinction #
    beta_mol = molecular_backscatter(
        wavelength, pressure, temperature, component=component
    )
    alfa_mol = molecular_extinction(wavelength, pressure, temperature)
    lr_mol = molecular_lidar_ratio(wavelength, component=component)

    depo_mol = molecular_depolarization(wavelength, component=component)

    """ transmittance """
    transmittance_array = transmittance(alfa_mol, heights)

    """ attenuated molecular backscatter """
    att_beta_mol = attenuated_backscatter(beta_mol, transmittance_array)

    if times is None:
        mol_properties = xr.Dataset(
            {
                "molecular_beta": (["range"], beta_mol),
                "molecular_alpha": (["range"], alfa_mol),
                "attenuated_molecular_beta": (["range"], att_beta_mol),
                "molecular_lidar_ratio": lr_mol,
                "molecular_depolarization": depo_mol,
            },
            coords={"range": heights},
        )
    else:
        mol_properties = xr.Dataset(
            {
                "molecular_beta": (["time", "range"], beta_mol),
                "molecular_alpha": (["time", "range"], alfa_mol),
                "attenuated_molecular_beta": (["time", "range"], att_beta_mol),
                "molecular_lidar_ratio": ([], lr_mol),
                "molecular_depolarization": depo_mol,
            },
            coords={
                "time": times,
                "range": heights,
            },  # FIXME: time is not used in the calculation yet
        )
    return mol_properties


def extrapolate_aod(
    wv1: float,
    wv2: float,
    aod2: np.ndarray[Any, np.dtype[np.float64]],
    ae: float,
) -> np.ndarray[Any, np.dtype[np.float64]]:
    """Extrapolates AOD at given wavelength using AOD at near wavelength and an appropriate Angstrom Exponent
    The Equation
    .. math::
        aod_{\\lambda_1} = aod_{\\lambda_2} \\cdot \\dfrac{\\lambda_1}{\\lambda_2} \\cdot AE(\\lambda_x - \\lambda_y)

    Args:
        wv1 (_type_): wavelength 1
        wv2 (_type_): wavelength 2
        aod2 (_type_): aerosol optical depth
        ae (_type_): Angstrom exponent coefficient

    Returns:
        np.ndarray[Any, np.dtype[np.float64]]: extrapolated AOD
    """

    return aod2 * (wv1 / wv2) ** (-ae)


def interpolate_aod(
    wv_arr: float, aod_arr: np.ndarray[Any, np.dtype[np.float64]], wv0: float
) -> np.ndarray[Any, np.dtype[np.float64]]:
    """
    Fit log_wv, log_aod to 2nd-order Polynomial [O'Neill et al., 2001]

    Args:
        wv_arr (_type_): wavelength
        aod_arr (_type_): aerosol optical depth
        wv0 (_type_): target wavelength

    Returns:
        np.ndarray[Any, np.dtype[np.float64]]: interpolated AOD
    """

    y = np.log(aod_arr)
    x = np.log(wv_arr)

    coeff = np.polyfit(x, y, 2, full=True)
    # rr = np.sum( (y - np.polyval(coeff[0], x)) **2) # residuals
    if coeff[1][0].squeeze() < 0.1:
        aod0 = np.exp(np.polyval(coeff[0], np.log(wv0)))
    else:
        raise ValueError("Fit not appropiate.")

    return aod0


def calculate_angstrom_exponent(
    wv1: float, wv2: float, aod1: float | np.ndarray, aod2: float | np.ndarray
) -> float | np.ndarray[Any, np.dtype[np.float64]]:
    """Retrieve Ansgstrom exponent using:
    .. math::
        $aod_{\\lambda_1} = aod_{\\lambda_2} \\cdot \\dfrac{\\lambda_1}{\\lambda_2} \\cdot AE(\\lambda_x - \\lambda_y)$

    Args:
        wv1 (float): first wavelength
        wv2 (float): second wavelength
        aod1 (float | np.ndarray): AOD at first wavelength
        aod2 (float | np.ndarray): AOD at second wavelength

    Returns:
        float | np.ndarray: angstrom exponent coefficient
    """
    return -(np.log(aod1 / aod2)) / (np.log(wv1 / wv2))



import sys
import pkg_resources

import typer
from loguru import logger

from .lidar.main import app as lidar_app

app = typer.Typer(
    name="GFATPy CLI",
    help="A console interface for the GFATPy utilities.",
    no_args_is_help=True,
)

app.add_typer(
    lidar_app,
    name="lidar",
    help="Lidar utilities to convert, read and plot lidar data.",
)


def version_callback(value: bool) -> None:
    if value:
        version = pkg_resources.get_distribution("gfatpy").version
        typer.echo(version)
        raise typer.Exit()


@app.callback()
def main(
    version: bool = typer.Option(
        False,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Shows gfatpy version information",
    ),
    verbose: bool = typer.Option(False, "--verbose"),
):
    if verbose:
        logger.remove()
        logger.add(sys.stdout, level="DEBUG")



# [cómo usar restructuretext --> https://www.sphinx-doc.org/es/master/usage/restructuredtext/index.html]

from .lidar_preprocessing import preprocess
from . import gluing_de_la_rosa, lidar_gluing_bravo_aranda

__all__ = ["preprocess", "gluing_de_la_rosa", "lidar_gluing_bravo_aranda"]

__doc__ = """
This module contains a file `gfatpy.lidar.preprocessing.preprocess` which apply the following corrections to the lidar raw data:

* Dark measurement [AN only]
* Dead time correction [PC only]
* Trigger delay [PC only] [tbi]
* Zero bin [AN only]
* Background substraction
* Merge of polarized channels

The following files contains the depolarization calibation factors and the GHK parameters:
* `gfatpy.lidar.preprocessing.depolarization_calibration_mhc`
* `gfatpy.lidar.preprocessing.depolarization_calibration_alh`
* `gfatpy.lidar.preprocessing.depolarization_calibration_vlt`

.. warning::
   Be sure these files are updated.

### Gluing:

There are some gluing implementations already. The main two are:
- `gfatpy.lidar.preprocessing.gluing_de_la_rosa`: Calculates a windowed correlation coefficient in everypoint within a range.
   Then, selects the highest and with a linear regression scales analog channel to photocounting untis.
   It works, but needs consistency.  For a good adjustment, windows has to be large and it works even better if the adjustment window is even larger. For instance:
   ```
   win_an = rcs_an_whole[idx, gl_bin - str(hw * 1.5) : gl_bin + hw * 5]
   win_pc = rcs_pc_whole[idx, gl_bin - str(hw * 1.5) : gl_bin + hw * 5]
   ```
   For instance in the default configuration The day /2021/07/05 from MULHACEN has a gluing inconsistency at time profile 217.
   Neighbours are a bit higher, resulting in a visible vertical line in the quicklook.
   It is thought that it happens because with larger adjustment windows the regression captures better how the scale (a and b params of the linear regression)
   despite the noise at higher altitudes.

- `gfatpy.lidar.preprocessing.lidar_gluing_bravo_aranda`: Calculates a windowed linear regression in everypoing for both channels and within a range.
   Chooses the point where this is minimum  and make a gluing with another window larger than the linear regression.
   For larger adjustment windows it works better.

"""



from . import preprocessing, retrieval, quality_assurance, plot, utils, nc_convert


__all__ = [
    "preprocessing",
    "retrieval",
    "quality_assurance",
    "plot",
    "utils",
    "nc_convert",
]


__doc__ = """
    The top of the lidar module that is compatible (at least) with GFAT lidars: "Veleta", "Mulhacén" and "Alhambra"
"""



import pandas as pd
import datetime as dt


def create_calibration_dataframe_vlt():
    # Dictionary organized by wavelengths
    depoCalib = {"xf": {}}

    depoCalib["xf"]["355"] = pd.DataFrame(
        columns=["date", "eta_an", "eta_pc", "GR", "GT", "HR", "HT", "K"]
    )

    dict0 = {
        "date": dt.datetime(2021, 12, 1),
        "eta_an": 1.0,
        "eta_pc": 1.0,
        "GR": 1.0,
        "GT": 1,
        "HR": -1.0,
        "HT": 1.0,
        "K": 1.0,
    }  # Considering y = + 1 i.e., parallel in T.

    dict1 = {
        "date": dt.datetime(2021, 12, 1),
        "eta_an": 1.0,
        "eta_pc": 1.0,
        "GR": 1.0,
        "GT": 1,
        "HR": 1.0,
        "HT": -1.0,
        "K": 1.0,
    }  # Considering y = - 1 i.e., parallel in R.
    depoCalib["xf"]["355"].loc[0] = pd.Series(dict0)
    depoCalib["xf"]["355"] = depoCalib["xf"]["355"].set_index("date")
    return depoCalib



import datetime as dt
from pathlib import Path
from typing import Tuple
import numpy as np
import pandas as pd
import xarray as xr
from scipy.signal import savgol_filter

from gfatpy.utils import utils
from gfatpy.atmo import atmo
from gfatpy.lidar.preprocessing import preprocess
from gfatpy.lidar.file_manager import channel2info
from gfatpy.lidar.utils import LIDAR_INFO, signal_to_rcs
from gfatpy.lidar.quality_assurance.io import rayleigh2earlinet, get_meteo
from gfatpy.lidar.quality_assurance.plot import plot_rayleigh_fit

lidar_id = "gr"
lidar_location = "Granada"

""" Types of Lidar Measurements """
measurement_types = {
    "RS": "Prs",
    "DC": "Pdc",
    "OT": "Pot",
    "DP-P45": "Pdp-45",
    "DP-N45": "Pdp-N45",
}


def rayleigh_fit_channel(
    channel: str,
    rf_dataset: xr.Dataset,
    initial_date: dt.datetime,
    final_date: dt.datetime,
    meteo_profiles: pd.DataFrame,
    meteo_info: dict,
    reference_range: Tuple[float, float],
    smooth_window: float,
) -> xr.Dataset:

    wavelength, *_ = channel2info(channel)

    # Molecular Attenuated Backscatter
    temperature = np.array(meteo_profiles["temperature"])
    pressure = np.array(meteo_profiles["pressure"])
    range = np.array(meteo_profiles["height"])
    mol_properties = atmo.molecular_properties(wavelength, pressure, temperature, range)
    att_beta_mol = mol_properties["attenuated_molecular_beta"]

    wavelength, _, polarization, mode = channel2info(channel)

    # RCS
    signal = rf_dataset[f"signal_{channel}"]
    rcs = signal_to_rcs(signal, signal.range)

    # Smooth rcs
    """ Lidar Resolution, Smoothing Bins """
    resolution = np.median(np.diff(signal.range))
    smooth_bins = np.round((smooth_window / resolution)).astype(int)
    sm_rcs = xr.DataArray(
        savgol_filter(rcs, smooth_bins, 3),
        coords={"range": signal.range},
        dims=["range"],
    )

    # Normalize time-averaged RCS, BCS nd Smoothed Time-Averaged RCS
    n_rcs = rcs / rcs.sel(range=slice(*reference_range)).mean("range")

    n_sm_rcs = sm_rcs / sm_rcs.sel(range=slice(*reference_range)).mean("range")

    n_att_beta_mol = att_beta_mol / att_beta_mol.sel(
        range=slice(*reference_range)
    ).mean("range")

    # output dataset will have height info in km
    ranges_km = range * 1e-3
    z_min_km = reference_range[0] * 1e-3
    z_max_km = reference_range[1] * 1e-3

    wavelength = xr.DataArray(data=wavelength, attrs={"str": f"{wavelength}"})
    polarization_long_name = LIDAR_INFO["metadata"]["code_polarization_str2long_name"][polarization]  # type: ignore
    polarization = xr.DataArray(
        data=polarization,
        dims=[],
        attrs={
            "long_name": polarization_long_name,
            "id": LIDAR_INFO["metadata"]["code_polarization_number2str"],
        },
    )
    polarization_code = LIDAR_INFO["metadata"]["code_mode_str2long_name"][mode]  # type: ignore
    detection_mode = xr.DataArray(
        data=mode,
        dims=[],
        attrs={
            "long_name": polarization_code,
            "id": LIDAR_INFO["metadata"]["code_mode_str2number"][mode],
        },
    )

    rcs = xr.DataArray(
        data=rcs,
        dims=["range"],
        coords={"range": ranges_km},
        attrs={
            "name": "RangeCorrectedSignal",
            "long_name": "range corrected signal avg",
            "units": "a.u.",
        },
    )

    n_rcs = xr.DataArray(
        data=n_rcs,
        dims=["range"],
        coords={"range": ranges_km},
        attrs={
            "name": "RangeCorrectedSignal",
            "long_name": "normalized- range-corrected signal.",
            "units": "a.u.",
        },
    )

    smoothed_rcs = xr.DataArray(
        data=sm_rcs,
        dims=["range"],
        coords={"range": ranges_km},
        attrs={
            "name": "RangeCorrectedSignal",
            "long_name": "smoothed range corrected signal",
            "units": "a.u.",
        },
    )

    normalized_smoothed_rcs = xr.DataArray(
        data=n_sm_rcs,
        dims=["range"],
        coords={"range": ranges_km},
        attrs={
            "name": "RangeCorrectedSignal",
            "long_name": "normalized- smoothed- range-corrected signal.",
            "units": "a.u.",
        },
    )

    attenuated_molecular_backscatter = xr.DataArray(
        data=att_beta_mol,
        dims=["range"],
        coords={"range": ranges_km},
        attrs={
            "name": "attnRayleighBSC",
            "long_name": "attenuated molecular backscatter",
            "units": "a.u.",
        },
    )

    normalized_attenuated_molecular_backscatter = xr.DataArray(
        data=n_att_beta_mol,
        dims=["range"],
        coords={"range": ranges_km},
        attrs={
            "name": "attnRayleighBSC",
            "long_name": "attenuated molecular backscatter norm",
            "units": "a.u.",
        },
    )
    dataset = xr.Dataset(
        data_vars={
            "wavelength": wavelength,
            "detection_mode": detection_mode,
            "RCS": rcs,
            "RCS_smooth": smoothed_rcs,
            "RCS_norm": n_rcs,
            "RCS_smooth_norm": normalized_smoothed_rcs,
            "BCS": attenuated_molecular_backscatter,
            "BCS_norm": normalized_attenuated_molecular_backscatter,
        },
        coords={"range": ranges_km},
        attrs={
            "lidar_location": lidar_location,
            "lidar_id": lidar_id,
            "lidar_name": rf_dataset.attrs["system"],
            "channel": channel,
            "radiosonde_location": meteo_info["radiosonde_location"],
            "radiosonde_wmo_id": meteo_info["radiosonde_wmo_id"],
            "radiosonde_datetime": meteo_info["radiosonde_datetime"],
            "datetime_ini": initial_date.strftime(
                "%Y-%m-%dT%H:%M:%S"
            ),  # FIXME: Dejar como estaba o pasar a ISO 8601?. Estaba en formato 20220808T12, ahora en 2022-08-08T12:00:00
            "datetime_end": final_date.strftime("%Y-%m-%dT%H:%M:%S"),
            "datetime_format": "%Y-%m-%dT%H:%M:%S",
            # "timestamp": final_date - initial_date,
            "duration": (final_date - initial_date).total_seconds(),
            "duration_units": "seconds",
            "rayleigh_height_limits": [z_min_km, z_max_km],
        },
    )
    dataset["range"].attrs["units"] = "km"
    dataset["range"].attrs["long_name"] = "height"

    if channel[-1] == "a" and f"dc_{channel}" in [*rf_dataset.variables.keys()]:
        dc_signal = rf_dataset[f"dc_{channel}"]
        # Normalized dc
        n_dc_signal = dc_signal / dc_signal.sel(range=slice(*reference_range)).mean(
            "range"
        )

        dataset["DC"] = xr.DataArray(
            data=dc_signal.values,
            dims=["range"],
            coords={"range": ranges_km},
            attrs={
                "name": "D",
                "long_name": "dark current avg",
                "units": "a.u.",
            },
        )
        dataset["DC_norm"] = xr.DataArray(
            data=n_dc_signal.values,
            dims=["range"],
            coords={"range": ranges_km},
            attrs={
                "name": "D",
                "long_name": "dark current avg norm",
                "units": "a.u.",
            },
        )

        dataset.attrs["dark_subtracted"] = "dark-subtracted"

    return dataset


def rayleigh_fit_from_file(
    file: Path,
    channels: list[str] | None = None,
    initial_hour: int | None = None,
    duration: float = 30,
    reference_range: Tuple[float, float] = (7000, 8000),
    smooth_window: float = 250,
    crop_ranges: tuple[float, float] = (0, 30000),
    range_limits: tuple[float, float] | None = None,
    meteorology_source: str = "standard_atmosphere",
    pressure_profile: np.ndarray
    | list
    | None = [
        3,
        5,
    ],  # TODO: Implementar definiciones por usuario
    temperature_profile: np.ndarray
    | list
    | None = [
        3,
        5,
    ],  # TODO: Implementar definiciones por usuario
    output_dir: Path | str | None = None,
    save_fig: bool = False,
):

    if output_dir is None:
        output_dir = Path.cwd()
    elif isinstance(output_dir, str):
        output_dir = Path(output_dir)

    if not output_dir.exists() or not output_dir.is_dir():
        raise NotADirectoryError(f"Output directory {output_dir} does not exist.")

    # Lidar preprocess
    lidar_ds = preprocess(
        file,
        channels=channels,
        crop_ranges=crop_ranges,
        save_dc=True,
        save_bg=True,
    )
    # time in array
    times = lidar_ds["time"].values
    times = np.array([utils.numpy_to_datetime(xx) for xx in times])

    # ranges in array
    range = lidar_ds["range"].values

    # Define initial and final date
    initial_date = dt.datetime(
        times[1].date().year, times[1].date().month, times[1].date().day
    )
    if initial_hour is None:
        initial_hour = (times[-1] - dt.timedelta(minutes=60)).hour
    if initial_hour is not None:
        initial_date = initial_date.replace(hour=initial_hour)
    else:
        raise ValueError("initial_hour not found.")
    final_date = initial_date + dt.timedelta(minutes=duration)

    # Select of period:
    rf = lidar_ds.sel(time=slice(initial_date, final_date)).mean("time")
    rf.attrs = lidar_ds.attrs

    # Get meteo profiles
    meteo_profiles, meteo_info = get_meteo(initial_date, range, meteorology_source)

    if channels is None:
        channels_: list[str] = rf.channel.values.tolist()
        channels = channels_

    # For channel
    for channel_ in channels:
        if channel_ in rf.channel.values:
            dataset = rayleigh_fit_channel(
                channel_,
                rf,
                initial_date,
                final_date,
                meteo_profiles,
                meteo_info,
                reference_range,
                smooth_window,
            )

            file_nc, _ = rayleigh2earlinet(dataset, output_dir=output_dir)

            if save_fig:
                plot_rayleigh_fit(
                    file_nc, output_dir=output_dir, range_limits=range_limits
                )



from typing import Any, Tuple
import numpy as np
from scipy import integrate
from scipy.integrate import cumtrapz

from loguru import logger

from gfatpy.lidar.utils import refill_overlap


def klett_rcs(
    rcs_profile: np.ndarray,
    range_profile: np.ndarray,
    beta_mol_profile: np.ndarray,
    lr_aer: float = 45,
    lr_mol: float = 8 * np.pi / 3,
    reference: Tuple[float, float] = (8000, 8500),
    beta_aer_ref: float = 0,
) -> np.ndarray:
    """Calculate aerosol backscattering using Classical Klett algorithm verified with Fernald, F. G.: Appl. Opt., 23, 652-653, 1984.

    Args:
        rcs_profile (np.ndarray): 1D signal profile.
        range_profile (np.ndarray): 1D range profile with the same shape as rcs_profile.
        beta_mol_profile (np.ndarray): 1D array containing molecular backscatter values.
        lr_mol (float): Molecular lidar ratio (default value based on Rayleigh scattering).
        lr_aer (float, optional): Aerosol lidar ratio (default is 45 sr).
        reference (Tuple[float, float], optional): Range interval (ymin and ymax) for reference calculation. Defaults to (8000, 8500).
        beta_aer_ref (float, optional): Aerosol backscatter at reference range (ymin and ymax). Defaults to 0.

    Returns:
        np.ndarray: Aerosol backscattering (beta) profile.
    """
    ymin, ymax = reference

    particle_beta = np.zeros(len(range_profile))

    ytop = np.abs(range_profile - ymax).argmin()

    range_resolution = np.median(np.diff(range_profile))

    idx_ref = np.logical_and(range_profile >= ymin, range_profile <= ymax)

    if not idx_ref.any():
        raise ValueError("Range [ymin, ymax] out of rcs size.")

    calib = np.nanmean(
        rcs_profile[idx_ref] / (beta_mol_profile[idx_ref] + beta_aer_ref)
    )

    integer1 = np.flip(-cumtrapz(np.flip(beta_mol_profile[:ytop]), dx=range_resolution, initial=0))
    integrando = rcs_profile[:ytop] * np.exp(-2 * (lr_aer - lr_mol) * integer1)
    integer3 = np.flip(-cumtrapz(np.flip(integrando), dx=range_resolution, initial=0))

    particle_beta[:ytop] = (rcs_profile[:ytop] * np.exp(-2 * (lr_aer - lr_mol) * integer1)) / (calib - 2 * lr_aer * integer3) - beta_mol_profile[:ytop]

    return particle_beta


def klett_likely_bins(
    rcs_profile: np.ndarray[Any, np.dtype[np.float64]],
    att_mol_beta: np.ndarray[Any, np.dtype[np.float64]],
    heights: np.ndarray[Any, np.dtype[np.float64]],
    min_height: float = 1000,
    max_height: float = 1010,
    window_size: int = 50,
    step: int = 1,
):
    window_size // 2
    i_bin, e_bin = np.searchsorted(heights, [min_height, max_height])

    for i in np.arange(i_bin, e_bin + 1):
        rcs_profile / rcs_profile

    # return rcs_profile


def find_lidar_ratio(
    rcs: np.ndarray[Any, np.dtype[np.float64]],
    height: np.ndarray[Any, np.dtype[np.float64]],
    beta_mol: np.ndarray[Any, np.dtype[np.float64]],
    lr_mol: float,
    reference_aod: float,
    mininum_height: float = 0,
    lr_initial: float = 50,
    lr_resol: float = 1,
    max_iterations: int = 100,
    rel_diff_aod_percentage_threshold: float = 1,
    debugging: bool = False,
    klett_reference: Tuple[float, float] = (7000, 8000),
) -> Tuple[float, float | None, bool]:
    """Iterative process to find the lidar ratio (lr) that minimizes the difference between the measured and the calculated aerosol optical depth (aod).

    Args:
        rcs (np.ndarray[Any, np.dtype[np.float64]]): Range Corrected Signal
        height (np.ndarray[Any, np.dtype[np.float64]]): Range profile
        beta_mol (np.ndarray[Any, np.dtype[np.float64]]): Molecular backscattering coefficient profile
        lr_mol (float): Molecular lidar ratio
        reference_aod (float): Reference aerosol optical depth
        mininum_height (float, optional): Fullover height. Defaults to 0.
        lr_initial (float, optional): _description_. Defaults to 50.
        lr_resol (float, optional): _description_. Defaults to 1.
        max_iterations (int, optional): _description_. Defaults to 100.
        rel_diff_aod_percentage_threshold (float, optional): _description_. Defaults to 1.
        debugging (bool, optional): _description_. Defaults to False.
        klett_reference (Tuple[float, float], optional): _description_. Defaults to (7000, 8000).

    Returns:
        Tuple[float, float | None, bool]: _description_
    """

    # Calculate range resolution
    range_resolution = np.median(np.diff(height)).item()

    # Initialize loop
    lr_, iter_, run, success = lr_initial, 0, True, False
    rel_diff_aod = None

    while run:
        iter_ = iter_ + 1

        # Calculate aerosol backscatter
        beta_ = klett_rcs(
            rcs, height, beta_mol, lr_aer=lr_, lr_mol=lr_mol, reference=klett_reference
        )

        # Refill beta profile from minimum height to surface to avoid overlap influence
        beta_ = refill_overlap(beta_, height)

        # Calculate aerosol optical depth
        aod_ = integrate.simps(beta_ * lr_, dx=range_resolution)

        # Calculate relative difference between measured and calculated aod
        rel_diff_aod = 100 * (aod_ - reference_aod) / reference_aod

        if debugging:
            print(
                "lidar_aod: %.3f| reference_aod: %.3f | relative_difference: %.1f%%"
                % (aod_, reference_aod, rel_diff_aod)
            )

        # Check convergence
        if np.abs(rel_diff_aod) > rel_diff_aod_percentage_threshold:
            if rel_diff_aod > 0:
                if lr_ < 10:
                    run = False
                    print("No convergence. LR goes too low.")
                else:
                    lr_ = lr_ - 1
            else:
                if lr_ > 150:
                    run = False
                    print("No convergence. LR goes too high.")
                else:
                    lr_ = lr_ + 1
        else:
            print("LR found: %f" % lr_)
            run = False
            success = True

        # Check maximum number of iterations
        if iter_ == max_iterations:
            run = False
            print("No convergence. Too many iterations.")

    return lr_, rel_diff_aod, success



"""
function needed to create plot with GFAT formatting
"""

import os
from pathlib import Path
from cycler import cycler

import matplotlib as mpl

# mpl.use('Agg')
import numpy as np
import datetime as dt
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib.axes import Axes
from matplotlib.offsetbox import AnchoredOffsetbox, OffsetImage


BASE_DIR = os.path.dirname(__file__)

COEFF = 2.0


COLORS = [
    "#1f77b4",
    "#ff7f0e",
    "#2ca02c",
    "#d62728",
    "#9467bd",
    "#8c564b",
    "#e377c2",
    "#7f7f7f",
    "#bcbd22",
    "#17becf",
]

plt.rcParams["axes.prop_cycle"] = cycler("color", COLORS)

plt.rcParams["xtick.major.pad"] = 1.5 * COEFF
plt.rcParams["xtick.minor.pad"] = 1.5 * COEFF
plt.rcParams["ytick.major.pad"] = 1.5 * COEFF
plt.rcParams["ytick.minor.pad"] = 1.5 * COEFF

plt.rcParams["xtick.major.size"] = 1.0 * COEFF
plt.rcParams["xtick.minor.size"] = 1.0 * COEFF
plt.rcParams["ytick.major.size"] = 1.0 * COEFF
plt.rcParams["ytick.minor.size"] = 1.0 * COEFF

plt.rcParams["xtick.labelsize"] = 5 * COEFF
plt.rcParams["ytick.labelsize"] = 5 * COEFF

plt.rcParams["axes.linewidth"] = 0.5 * COEFF
plt.rcParams["axes.labelsize"] = 5 * COEFF
plt.rcParams["axes.facecolor"] = "#c7c7c7"

plt.rcParams["legend.numpoints"] = 3
plt.rcParams["legend.fontsize"] = 3.5 * COEFF
plt.rcParams["legend.facecolor"] = "#ffffff"

plt.rcParams["grid.linestyle"] = ":"
plt.rcParams["grid.linewidth"] = 0.5 * COEFF
plt.rcParams["grid.alpha"] = 0.5


plt.rcParams["figure.subplot.hspace"] = 0.2
plt.rcParams["figure.subplot.wspace"] = 0.2
plt.rcParams["figure.subplot.bottom"] = 0.11
plt.rcParams["figure.subplot.left"] = 0.14
plt.rcParams["figure.subplot.right"] = 0.95
plt.rcParams["figure.subplot.top"] = 0.82

plt.rcParams["figure.figsize"] = 2.913 * COEFF, 2.047 * COEFF
plt.rcParams["figure.facecolor"] = "#ffffff"

plt.rcParams["lines.markersize"] = 2.6 * COEFF
plt.rcParams["lines.markeredgewidth"] = 0.5 * COEFF
plt.rcParams["lines.linewidth"] = 0.5 * COEFF


def title1(mytitle, coef):
    """
    inclus le titre au document.
        @param mytitle: titre du document.
        @param coef : coefficient GFAT (renvoye par la fonction formatGFAT).
    """

    plt.figtext(
        0.5,
        0.95,
        mytitle,
        fontsize=6.5 * coef,
        fontweight="bold",
        horizontalalignment="center",
        verticalalignment="center",
    )
    return


def title2(mytitle, coef):
    """
    inclus le sous titre au document.
        @param mytitle: titre du document.
        @param coef : coefficient GFAT (renvoye par la fonction formatGFAT).
    """

    plt.figtext(
        0.5,
        0.89,
        mytitle,
        fontsize=5.5 * coef,
        horizontalalignment="center",
        verticalalignment="center",
    )
    return


def title3(mytitle, coef):
    """
    inclus le sous sous titre au document.
        @param mytitle: titre du document.
        @param coef : coefficient GFAT (renvoye par la fonction formatGFAT).
    """
    plt.figtext(
        0.5,
        0.85,
        mytitle,
        fontsize=4.5 * coef,
        horizontalalignment="center",
        verticalalignment="center",
    )
    return


def watermark(
    ax: Axes | np.ndarray = plt.gca(),
    zoom: float = 0.5,
    alpha=0.25,
    xpos=65,
    ypos=315,
    logofile="GFAT",
) -> None:
    """Place watermark in bottom right of figure.
    fig: figure handle
    ax: axes handle
    alpha: alpha channel, ie transparency
    xpos: horizontal location of the figure in pixel
    ypos: vertical location of the figure in pixel
    logofile: file path of the image to use as logo. Default: 'GFAT' redirects to the GFAT logo.
    """

    # Get the pixel dimensions of the figure
    # Import logo and scale accordingly
    if logofile == "GFAT":
        file_path = Path(__file__).parent.parent / "assets" / "LOGO_GFAT_150pp.png"
    else:
        raise NotImplementedError(f"{logofile} name not implemented")

    img = plt.imread(file_path)  # type: ignore
    image_box = OffsetImage(img, alpha=alpha, zoom=zoom)
    ao = AnchoredOffsetbox(
        "upper left", pad=0, borderpad=0.5, child=image_box, frameon=False
    )

    if isinstance(ax, Axes):
        ax.add_artist(ao)
    elif isinstance(ax, np.ndarray):
        _ax: Axes
        for _ax in ax.flat:
            _ax.add_artist(ao)

    # wm_width = int(width / scale)  # make the watermark 1/10 of the figure size
    # scaling = wm_width / float(img.size[0])
    # wm_height = int(float(img.size[1]) * float(scaling))
    # img = img.resize((wm_width, wm_height), Image.Resampling.LANCZOS)

    # # Place the watermark in the lower right of the figure
    # plt.figimage(img, xpos, ypos, alpha=alpha, zorder=1)


def tick():
    matplotlib.ticker.FuncFormatter(tmp_f)


def tmp_f(date_dt):
    """
    convert datetime to numerical date
    """
    return mdates.num2date(date_dt).strftime("%H")


def color_list(n):
    import matplotlib.pylab as pl

    colors = pl.cm.jet(np.linspace(0, 1, n))
    return colors


def font_axes(ax, fontsize=14):
    for item in (
        [ax.title, ax.xaxis.label, ax.yaxis.label]
        + ax.get_xticklabels()
        + ax.get_yticklabels()
    ):
        item.set_fontsize(fontsize)


def gapsizer(ax, time, range, gapsize, colour="#c7c7c7"):
    """
    This function creates a rectangle of color 'colour' when time gap
    are found in the array 'time'.
    """
    # search for holes in data
    # --------------------------------------------------------------------
    dif_time = time[1:] - time[0:-1]
    print(type(dif_time))
    for index, delta in enumerate(dif_time):
        if delta > dt.timedelta(minutes=gapsize):
            # missing hide bad data
            start = mdates.date2num(time[index])
            end = mdates.date2num(time[index + 1])
            width = end - start

            # Plot rectangle
            end = mdates.date2num(time[index + 1])
            rect = mpl.patches.Rectangle(
                (start, 0), width, np.nanmax(range), color=colour
            )
            ax.add_patch(rect)


class OOMFormatter(mpl.ticker.ScalarFormatter):
    def __init__(self, order=0, fformat="%1.1f", offset=True, mathText=True):
        self.oom = order
        self.fformat = fformat
        mpl.ticker.ScalarFormatter.__init__(
            self, useOffset=offset, useMathText=mathText
        )

    def _set_orderOfMagnitude(self, nothing):
        self.orderOfMagnitude = self.oom

    def _set_format(self, vmin, vmax):
        self.format = self.fformat
        if self._useMathText:
            self.format = "$%s$" % mpl.ticker._mathdefault(self.format)



# This Python script will be executed from within the main lidar_correction_ghk.py
# Probably it will be better in the future to let the main script rather read a conguration file,
# which might improve the portability of the code within an executable.
# Due to problems I had with some two letter variables, most variables are now with at least
# three letters mixed small and capital.

# Do you want to calculate the errors? If not, just the GHK-parameters are determined.
Error_Calc = False

# Header to identify the lidar system
#
EID = "gr"  # Earlinet station ID
LID = "mulhacen"  # Additional lidar ID (short descriptive text)
print("Lidar system :", EID, ", ", LID)

# +++ IL Laser and +-Uncertainty
bL = 1.0  # degree of linear polarization; default 1
Qin, dQin, nQin = (
    1.0,
    0.0,
    0,
)  # second Stokes vector parameter; default 1 => linear polarization  0.999 => LDR = 0.0005
Vin, dVin, nVin = (
    0.0,
    0.0,
    0,
)  # fourth Stokes vector parameter; default 0  => corresponds to LDR 0.0005 with DOP 1
RotL, dRotL, nRotL = (
    alpha_value,
    alpha_error,
    1,
)  # alpha; rotation of laser polarization in degrees; default 0

# +++ ME Emitter optics and +-Uncertainty;  default = no emitter optics
DiE, dDiE, nDiE = 0.0, 0.1, 1  # Diattenuation
TiE = 1.0  # Unpolarized transmittance
RetE, dRetE, nRetE = 0.0, 180.0, 1  # Retardance in degrees
RotE, dRotE, nRotE = 0.0, 1.0, 1  # beta: Rotation of optical element in degrees

# +++ MO Receiver optics including telescope
DiO, dDiO, nDiO = 0.0, 0.1, 1
TiO = 1.0
RetO, dRetO, nRetO = 0.0, 180.0, 1
RotO, dRotO, nRotO = 0.0, 0.5, 1  # gamma: Rotation of optical element in degrees

# +++++ PBS MT Transmitting path defined with TS, TP, PolFilter extinction ratio ERaT, and +-Uncertainty
#   --- Polarizing beam splitter transmitting path
TP, dTP, nTP = 0.95, 0.01, 1
TS, dTS, nTS = 0.005, 0.005, 1
RetT, dRetT, nRetT = 0.0, 180.0, 0  # Retardance in degrees
#   --- Pol.Filter behind transmitted path of PBS
ERaT, dERaT, nERaT = 1e-3, 5e-4, 1  # Extinction ratio
RotaT, dRotaT, nRotaT = (
    0.0,
    1.0,
    1,
)  # Rotation of the Pol.-filter in degrees; usually 0° because TP >> TS, but for PollyXTs it can also be 90°
#   --
TiT = 0.5 * (TP + TS)
DiT = (TP - TS) / (TP + TS)
DaT = (1 - ERaT) / (1 + ERaT)
TaT = 0.5 * (1 + ERaT)

# +++++ PBS MR Reflecting path defined with RS, RP, PolFilter extinction ratio ERaR  and +-Uncertainty
#   ---- for PBS without absorption the change of RS and RP must depend on the change of TP and TS. Hence the values and uncertainties are not independent.
RS_RP_depend_on_TS_TP = True
#   --- Polarizing beam splitter reflecting path
if RS_RP_depend_on_TS_TP:
    RP, dRP, nRP = 1 - TP, 0.00, 1  # do not change this
    RS, dRS, nRS = 1 - TS, 0.00, 1  # do not change this
else:
    RP, dRP, nRP = 0.05, 0.01, 0  # change this if RS_RP_depend_on_TS_TP = False
    RS, dRS, nRS = 0.98, 0.01, 0  # change this if RS_RP_depend_on_TS_TP = False
RetR, dRetR, nRetR = 0.0, 180.0, 0
#   --- Pol.Filter behind reflected path of PBS
ERaR, dERaR, nERaR = 1e-3, 5e-4, 1  # Extinction ratio
RotaR, dRotaR, nRotaR = (
    90.0,
    0.0,
    1,
)  # Rotation of the Pol.-filter in degrees; usually 90° because RS >> RP, but for PollyXTs it can also be 0°
#   --
TiR = 0.5 * (RP + RS)
DiR = (RP - RS) / (RP + RS)
DaR = (1 - ERaR) / (1 + ERaR)
TaR = 0.5 * (1 + ERaR)

# +++ Parallel signal detected in the transmitted channel => Y = +1, or in the reflected channel => Y = -1
Y = -1.0

# +++ Calibrator Location
LocC = 3  # location of calibrator: 1 = behind laser; 2 = behind emitter; 3 = before receiver; 4 = before PBS
# --- Calibrator Type used; defined by matrix values below
TypeC = 3  # Type of calibrator: 1 = mechanical rotator; 2 = hwp rotator (fixed retardation); 3 = linear polarizer; 4 = qwp; 5 = circular polarizer; 6 = real HWP calibration +-22.5°
# --- MC Calibrator parameters
if TypeC == 1:  # mechanical rotator
    DiC, dDiC, nDiC = 0.0, 0.0, 0
    TiC = 1.0
    RetC, dRetC, nRetC = 0.0, 0.0, 0
    RotC, dRotC, nRotC = -5.9, 0.3, 1  # constant calibrator offset epsilon
    # Rotation error without calibrator: if False, then epsilon = 0 for normal measurements
    RotationErrorEpsilonForNormalMeasurements = (
        True  # 	is in general True for TypeC == 1 calibrator
    )
elif TypeC == 2:  # HWP rotator
    DiC, dDiC, nDiC = 0.0, 0.0, 0
    TiC = 1.0
    RetC, dRetC, nRetC = 180.0, 0.0, 0
    # NOTE: use here twice the HWP-rotation-angle
    RotC, dRotC, nRotC = 0.0, 0.1, 1  # constant calibrator offset epsilon
    RotationErrorEpsilonForNormalMeasurements = (
        True  # 	is in general True for TypeC == 2 calibrator
    )
elif (
    TypeC == 3
):  # linear polarizer calibrator. Diattenuation DiC = (1-ERC)/(1+ERC); ERC = extinction ratio of calibrator
    DiC, dDiC, nDiC = 0.9998, 0.0001, 1  # ideal 1.0
    TiC = 0.4  # ideal 0.5
    RetC, dRetC, nRetC = 0.0, 0.0, 0
    RotC, dRotC, nRotC = (
        epsilon_value,
        epsilon_error,
        1,
    )  # constant calibrator offset epsilon
    RotationErrorEpsilonForNormalMeasurements = (
        True  # 	is in general False for TypeC == 3 calibrator
    )
elif TypeC == 4:  # QWP calibrator
    DiC, dDiC, nDiC = 0.0, 0.0, 0  # ideal 1.0
    TiC = 1.0  # ideal 0.5
    RetC, dRetC, nRetC = 90.0, 0.0, 0
    RotC, dRotC, nRotC = 0.0, 0.1, 1  # constant calibrator offset epsilon
    RotationErrorEpsilonForNormalMeasurements = (
        False  # 	is  False for TypeC == 4 calibrator
    )
elif (
    TypeC == 6
):  # real half-wave plate calibration at +-22.5°  => rotated_diattenuator_X22x5deg.odt
    DiC, dDiC, nDiC = 0.0, 0.0, 0
    TiC = 1.0
    RetC, dRetC, nRetC = 180.0, 0.0, 0
    # Note: use real HWP angles here
    RotC, dRotC, nRotC = 0.0, 0.1, 1  # constant calibrator offset epsilon
    RotationErrorEpsilonForNormalMeasurements = (
        True  # 	is in general True for TypeC == 6 calibrator
    )
else:
    print("calibrator not implemented yet")
    sys.exit()

# --- LDRCal assumed atmospheric linear depolarization ratio during the calibration measurements in calibration range with almost clean air (first guess)
LDRCal, dLDRCal, nLDRCal = 0.00355, 0.0005, 1  # spans the interference filter influence

# ====================================================
# NOTE: there is no need to change anything below.

# --- LDRtrue for simulation of measurement => LDRsim
LDRtrue = 0.4
LDRtrue2 = 0.004

# --- measured LDRm will be corrected with calculated parameters GHK
LDRmeas = 0.3

# --- this is just for correct transfer of the variables to the main file
RotL0, dRotL, nRotL = RotL, dRotL, nRotL
# Emitter
DiE0, dDiE, nDiE = DiE, dDiE, nDiE
RetE0, dRetE, nRetE = RetE, dRetE, nRetE
RotE0, dRotE, nRotE = RotE, dRotE, nRotE
# Receiver
DiO0, dDiO, nDiO = DiO, dDiO, nDiO
RetO0, dRetO, nRetO = RetO, dRetO, nRetO
RotO0, dRotO, nRotO = RotO, dRotO, nRotO
# Calibrator
DiC0, dDiC, nDiC = DiC, dDiC, nDiC
RetC0, dRetC, nRetC = RetC, dRetC, nRetC
RotC0, dRotC, nRotC = RotC, dRotC, nRotC
# PBS
TP0, dTP, nTP = TP, dTP, nTP
TS0, dTS, nTS = TS, dTS, nTS
RetT0, dRetT, nRetT = RetT, dRetT, nRetT

ERaT0, dERaT, nERaT = ERaT, dERaT, nERaT
RotaT0, dRotaT, nRotaT = RotaT, dRotaT, nRotaT

RP0, dRP, nRP = RP, dRP, nRP
RS0, dRS, nRS = RS, dRS, nRS
RetR0, dRetR, nRetR = RetR, dRetR, nRetR

ERaR0, dERaR, nERaR = ERaR, dERaR, nERaR
RotaR0, dRotaR, nRotaR = RotaR, dRotaR, nRotaR

LDRCal0, dLDRCal, nLDRCal = LDRCal, dLDRCal, nLDRCal



from atmospheric_lidar.licel import LicelLidarMeasurement
from . import alhambra_729


class AlhambraLidarMeasurement(LicelLidarMeasurement):
    extra_netcdf_parameters = alhambra_729

    def __init__(self, file_list=None, use_id_as_name=True):
        super(AlhambraLidarMeasurement, self).__init__(file_list, use_id_as_name)

    def set_PT(self):
        """Sets the pressure and temperature at station level .
        The results are stored in the info dictionary.
        """

        self.info["Temperature"] = 25.0
        self.info["Pressure"] = 1020.0



from datetime import datetime

import matplotlib.pyplot as plt
import numba
import numpy as np
import xarray as xr
from gfatpy.atmo.atmo import molecular_properties
from gfatpy.atmo.ecmwf import get_ecmwf_day
from gfatpy.utils.utils import moving_average, parse_datetime
from numba.typed import List
from scipy.signal import savgol_filter


def molecular_properties_2d(
    date: datetime | str,
    heights: np.ndarray,
    times: np.ndarray,
    wavelength: float = 532,
) -> xr.Dataset:
    """A function that request ECMWF temperatures and presures for a whole day.
    Then, pipes them to `gfatpy.atmo.atmo.molecular_properties`

    Args:
        date (datetime | str): date
        heights (np.ndarray): height ranges
        time (np.ndarray): time
        wavelength (float, optional): wavelength. Defaults to 532.

    Returns:
        xr.Dataset: it contains ["molecular_beta", "molecular_alpha", "attenuated_molecular_beta", "molecular_lidar_ratio"].
    """

    _date = parse_datetime(date)
    atmo_d = get_ecmwf_day(_date, heights=heights, times=times)

    mol_d = molecular_properties(
        wavelength,
        pressure=atmo_d.pressure.values,
        temperature=atmo_d.temperature.values,
        heights=atmo_d.range.values,
        times=atmo_d.time.values,
    )
    return mol_d


def iterative_fitting(
    rcs_profile: np.ndarray,
    attenuated_molecular_backscatter: np.ndarray,
    window_size_bins: int = 5,
    min_bin: int = 600,
    max_bin: int = 1000,
) -> np.ndarray:

    if rcs_profile.shape != attenuated_molecular_backscatter.shape:
        raise ValueError(f"RCS and Betta ranges must match")

    x_axis = np.arange(window_size_bins * 2)
    bool_matrix = np.full_like(rcs_profile, False, dtype=np.bool8)

    slope = []
    slope_mol = []

    for idx in np.arange(min_bin, max_bin):
        _rcs_norm = rcs_profile / rcs_profile[idx]
        _att_norm = (
            attenuated_molecular_backscatter / attenuated_molecular_backscatter[idx]
        )

        prof_slice = _rcs_norm[idx - window_size_bins : idx + window_size_bins]
        att_slice = _att_norm[idx - window_size_bins : idx + window_size_bins]

        coeff_prof = np.polyfit(np.arange(window_size_bins * 2), prof_slice, 1)
        coeff_att = np.polyfit(np.arange(window_size_bins * 2), att_slice, 1)

        slope.append(coeff_prof[0])
        slope_mol.append(coeff_att[0])

        # plt.scatter(x_axis, prof_slice, c="g")
        # plt.plot(x_axis, np.polyval(coeff_prof, x_axis), c="g")

        # plt.scatter(x_axis, att_slice, c="b")
        # plt.plot(x_axis, np.polyval(coeff_att, x_axis), c="b")

        # plt.show()

        att_data = np.polyval(coeff_att, x_axis)
        r2 = 1 - (
            ((prof_slice - att_data) ** 2).sum()
            / ((prof_slice - att_data.mean()) ** 2).sum()
        )

        # print(f'Mol m: {coeff_att[0]}')
        # print(f'Prof m: {coeff_prof[0]}')
        # if r2 > 0.25:
        #     print(f"R^2: {r2}")

    # plt.plot(slope)
    # plt.plot(slope_mol)
    # plt.close()
    return bool_matrix


def split_continous_measurements(
    time_array: np.ndarray, time_greater_than: float = 121 # Two minutest to avoid one profile search multiple DC
) -> list[np.ndarray]:
    """Groups times array into clusters with no more that `time_greater_than`

    Args:
        time_array (np.ndarray): Time

    Returns:
        list[np.ndarray]: list of lidar measurement slices
    """
    diffs = (time_array[1:] - time_array[0:-1]).astype("f") / 1e9  # type: ignore
    return np.split(time_array, np.where(diffs > time_greater_than)[0] + 1)


def mask_by_slope(
    rcs: xr.DataArray,
    att_beta: xr.DataArray,
    min_height: float = 4000,
    max_height: float = 7000,
    window_size: float = 1000,
    window_time: float = 30,
    max_rel_error: float = 0.15,
    plot_profile: int | None = None,
) -> xr.DataArray:
    rcs_sel = rcs.sel(range=slice(min_height, max_height))
    beta_sel = att_beta.sel(range=slice(min_height, max_height))

    rcs_sel = xr.apply_ufunc(
        moving_average, rcs_sel, kwargs={"window_size": window_time}, dask="allowed"
    )
    rcs_sel = xr.apply_ufunc(
        smooth_profiles, rcs_sel, kwargs={"window_size": 200}, dask="allowed"
    )

    result = np.full_like(beta_sel.values, False, dtype=bool)

    for idx, height in enumerate(rcs_sel.range):
        valid_ranges = rcs_sel.range[
            (height - window_size / 2 <= rcs_sel.range)
            & (height + window_size / 2 >= rcs_sel.range)
        ]
        height_rcs = rcs_sel.loc[dict(range=valid_ranges)]
        height_rcs /= height_rcs.isel(range=0)

        height_beta = beta_sel.loc[dict(range=valid_ranges)]
        height_beta /= height_beta.isel(range=0)

        x = np.arange(height_rcs.shape[1])

        slopes_rcs = np.polyfit(x, height_rcs.values.T, 1)[0]
        slopes_beta = np.polyfit(x, height_beta.values.T, 1)[0]

        result[:, idx] = (
            np.abs((slopes_rcs - slopes_beta) / slopes_beta) <= max_rel_error
        )

    result_data_array = xr.full_like(rcs, False)
    result_data_array.loc[dict(range=slice(min_height, max_height))] = result

    if plot_profile is not None:
        md = (max_height + min_height) / 2
        n_rcs = rcs[plot_profile] / rcs[plot_profile].sel(range=md, method="nearest")
        n_rcs_smth = rcs_sel[plot_profile] / rcs_sel[plot_profile].sel(
            range=md, method="nearest"
        )
        n_beta = att_beta[plot_profile] / att_beta[plot_profile].sel(
            range=md, method="nearest"
        )
        plt.plot(rcs.range, n_rcs)
        plt.plot(att_beta.range, n_beta)
        plt.plot(n_rcs_smth.range, n_rcs_smth)

        results = np.where(result_data_array[plot_profile], n_beta, np.nan)

        plt.plot(result_data_array.range, results, c="r", lw=2)

        plt.yscale("log")
        plt.show()

    return result_data_array


def mask_by_corrcoef(
    rcs: xr.DataArray,
    att_beta: xr.DataArray,
    min_height: float = 1000,
    max_height: float = 15000,
    window_size: float = 500,
    window_time: float = 15,
    min_corr: float = 0.95,
):

    rcs_sel = rcs.sel(range=slice(min_height, max_height))
    beta_sel = att_beta.sel(range=slice(min_height, max_height))

    rcs_sel = xr.apply_ufunc(smooth_profiles, rcs_sel, kwargs={"window_size": 170})
    rcs_sel = xr.apply_ufunc(
        moving_average, rcs_sel, kwargs={"window_size": window_time}
    )

    result = np.full_like(beta_sel.values, False, dtype=bool)
    ranges = rcs_sel.range.values

    lim_indexes = List()
    range_mask_limits = [
        np.where((h - window_size / 2 <= ranges) & (h + window_size / 2 >= ranges))[0]
        for h in ranges
    ]

    for idx in range(len(range_mask_limits)):
        c_list = List()
        c_list.append(range_mask_limits[idx][0])
        c_list.append(range_mask_limits[idx][-1])

        lim_indexes.append(c_list)

    result_array = windowed_correlation(
        rcs_sel.values,
        beta_sel.values,
        range_mask_limits=lim_indexes,
        min_corr=min_corr,
    )

    result_data_array = xr.full_like(rcs, False)
    result_data_array.loc[dict(range=slice(min_height, max_height))] = result_array

    return result_data_array
    # TODO:Finish function


def smooth_profiles(
    arr: np.ndarray, /, *, window_size: int = 11, polyorder: int = 3
) -> np.ndarray:
    def smooth_profile(_x: np.ndarray) -> np.ndarray:
        return savgol_filter(_x, window_size, 3)

    return np.apply_along_axis(smooth_profile, 1, arr)


@numba.njit(parallel=True)
def windowed_correlation(
    rcs: np.ndarray,
    att_beta: np.ndarray,
    /,
    *,
    range_mask_limits: list[list[int]],
    min_corr: float,
) -> np.ndarray:

    result_array = np.full(rcs.shape, np.nan)

    for t_idx in numba.prange(rcs.shape[0]):
        for r_idx in numba.prange(rcs.shape[1]):
            r_masks = range_mask_limits[r_idx]

            rcs_masked = rcs[t_idx][r_masks[0] : r_masks[1]]
            beta_masked = att_beta[t_idx][r_masks[0] : r_masks[1]]

            rcs_masked /= beta_masked[0]

            corr = np.corrcoef(rcs_masked, beta_masked)[0, 1]
            # set_trace()
            result_array[t_idx, r_idx] = corr >= min_corr

    return result_array


def cluster_value(arr: np.ndarray, /, *, value=1) -> list[list[tuple[int, int]]]:
    def cluster_1d(row) -> list[tuple[int, int]]:
        _i: int = 0
        count: int = 0
        clusters: list[tuple[int, int]] = []  # list[position, count]
        for (_i, *_), _v in np.ndenumerate(row):
            if _v == value and count == 0:
                count += 1
            elif _v == value and count != 0:
                count += 1
            elif _v != value and count == 0:
                continue
            elif _v != value and count != 0:
                clusters.append((_i - (1 + count), count))
                count = 0

        if count != 0:
            clusters.append((_i - (1 + count), count))

        return clusters

    time_cluster: list[list[tuple[int, int]]] = []

    for time_row in arr:
        time_cluster.append(cluster_1d(time_row))

    return time_cluster


def cluster_at_least(
    clusters: list[list[tuple[int, int]]], n_min: int = 5, /
) -> np.ndarray:
    result = np.full((len(clusters), 2), np.nan, dtype=float)

    for idx, prof_clust in enumerate(clusters):
        filtered = list(filter(lambda t: t[1] >= n_min, prof_clust))

        if len(filtered) == 0:
            continue

        selected = sorted(filtered, key=lambda t: t[1], reverse=True)[0]
        result[idx] = np.array([selected[0], selected[0] + selected[1]])

    return result




import numpy as np
from typing import Any, Tuple
from scipy.signal import savgol_filter

from gfatpy.utils import utils

""" LINEAR RESPONSE REGION
"""


def estimate_linear_response_region(
    ranges: np.ndarray[Any, np.dtype[np.float64]],
    an: np.ndarray[Any, np.dtype[np.float64]],
    pc: np.ndarray[Any, np.dtype[np.float64]],
    an_bg: np.ndarray[Any, np.dtype[np.float64]],
    an_bg_thres: float = 0.02,
    pc_thres: float = 15,
    region_range: Tuple = (1500, 10000),
    min_region: float = 1000,
    subregion: float = 500,
    n_step: int = 3,
    correlation_thres: float = 0.85,
):
    """

    Input:
    ranges: 1D array. ranges (m)
    an: analog signal preprocessed. 1D array.
    pc: photoncounting signal preprocessed 1D array.
    an_bg: background value for analog signal
    an_bg_thres = 0.1
    an_min: Min Thresh for AN
    pc_thres = 15MHz. Min Thresh. for PC
    region_range = (1000, 6000). Altitude Region to find linear response. tuple
    min_delta_h (m). length of range interval for LR region
    step_h (m). length of step for moving potential LR region
    r2_thres (>0, <1): threshold for R2 to consider LR
    """

    id_h = np.where(
        np.logical_and(ranges >= region_range[0], ranges <= region_range[1])
    )[0]
    ranges_aux = ranges[id_h]
    pc_aux = savgol_filter(pc[id_h], 21, polyorder=2)
    an_aux = savgol_filter(an[id_h], 21, polyorder=2)
    if np.logical_and(
        (pc_aux < np.float64(pc_thres)).any(),
        (an_aux < np.float64(an_bg * an_bg_thres)).any(),
    ):  #
        idx_min = 0
        while pc_aux[idx_min] > pc_thres:
            idx_min += 1
        idx_max = 0
        while an_aux[idx_max] > an_bg * an_bg_thres:
            idx_max += 1

        if ranges_aux[idx_max] - ranges_aux[idx_min] < min_region:
            idx_max = np.argmax(ranges_aux >= (ranges_aux[idx_min] + min_region))

        rvalue_arr = []
        slope_arr = []
        intercept_arr = []
        h_min_arr = []
        h_max_arr = []
        h_min = ranges[id_h][idx_min]
        h_max = ranges[id_h][idx_max]
        hi = h_min
        step = np.diff(ranges).mean()
        while hi < (h_max - subregion):
            he = hi + subregion
            while he < h_max:
                idx = np.logical_and(ranges >= hi, ranges <= he)
                a, b, rvalue_i = utils.linear_regression(an[idx], pc[idx])
                if rvalue_i >= correlation_thres:
                    slope_arr.append(a)
                    intercept_arr.append(b)
                    rvalue_arr.append(rvalue_i)
                    h_min_arr.append(ranges[idx][0])
                    h_max_arr.append(ranges[idx][-1])
                he += n_step * step
            hi += n_step * step
        if len(rvalue_arr) > 0:
            linear_region = True
            idx_lr = np.nanargmax(np.asarray(rvalue_arr))
            slope = slope_arr[idx_lr]
            intercept = intercept_arr[idx_lr]
            r2 = rvalue_arr[idx_lr]
            h_lr_min = h_min_arr[idx_lr]
            h_lr_max = h_max_arr[idx_lr]
        else:
            linear_region = False
            idx_lr, slope, intercept, r2, h_lr_min, h_lr_max = (
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
            )
    else:
        linear_region = False
        idx_lr, slope, intercept, r2, h_lr_min, h_lr_max = (
            np.nan,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
        )

    return h_lr_min, h_lr_max, r2, slope, intercept, linear_region
